def _binomial(n : int, m: int):
    if m == 0:
        return 1
    elif m > n / 2:
        return _binomial(n, n - m)
    else:
        return (n / m) * _binomial(n - 1, m - 1)

def binomial(n : int, m: int):
    if not isinstance(n, int):
        raise ValueError("n must be integer.")
    if not isinstance(m, int):
        raise ValueError("m must be integer.")
    if m < 0 or m > n:
        raise ValueError("m must be between 0 and n.")
    return _binomial(n, m)



import numpy as np
from typing import Tuple, List, Callable, Self
from numbers import Number
from pybezier.binomial import binomial

class BezierCurve(object):

    def __init__(self, points : np.ndarray, initial_time : float = 0, final_time : float = 1):
        if initial_time >= final_time:
            raise ValueError("Initial time must be smaller than final time.")
        self.points = points
        self.degree, self.dimension = points.shape
        self.degree -= 1
        self.initial_time = initial_time
        self.final_time = final_time
        self.duration = final_time - initial_time

    def initial_point(self) -> np.ndarray:
        return self.points[0]

    def final_point(self) -> np.ndarray:
        return self.points[-1]

    def _berstein(self, time : float | List[float], n : int) -> float:
        c = binomial(self.degree, n)
        t = (time - self.initial_time) / self.duration 
        value = c * t ** n * (1 - t) ** (self.degree - n)
        return value
    
    def _assert_same_times(self, curve : Self):
        if not np.isclose(self.initial_time, curve.initial_time):
            raise ValueError("Incompatible curves, initial times don't match.")
        if not np.isclose(self.final_time, curve.final_time):
            raise ValueError("Incompatible curves, final times don't match.")

    def __call__(self, time : float | List[float]) -> float:
        c = np.array([self._berstein(time, n) for n in range(self.degree + 1)])
        return c.T.dot(self.points)
    
    def __mul__(self, curve : Self | Number) -> Self:
        """See (44) in Algorithms for polynomials in Bernstein form, by Farouky
        and Rajan"""
        if not isinstance(curve, BezierCurve):
            curve = self._number_to_curve(curve)
        self._assert_same_times(curve)
        degree = self.degree + curve.degree
        dimension = max(self.dimension, curve.dimension)
        # dtype object is needed to handle curves whose control points
        # are variables in an optimization problem
        points = np.zeros((degree + 1, dimension), dtype=object)
        for i in range(degree + 1):
            j_min = max(0, i - curve.degree)
            j_max = min(self.degree, i)
            for j in range(j_min, j_max + 1):
                b = binomial(self.degree, j)
                b *= binomial(curve.degree, i - j)
                points[i] += self.points[j] * curve.points[i - j] * b
            points[i] /= binomial(degree, i)
        return BezierCurve(points, self.initial_time, self.final_time)
    
    def _number_to_curve(self, n : Number) -> Self:
        points = np.array([[n]])
        return BezierCurve(points, self.initial_time, self.final_time)
    
    def __rmul__(self, curve : Self | Number) -> Self:
        return self * curve

    def elevate_degree(self, degree : int) -> Self:
        points = np.ones((degree - self.degree + 1, 1))
        curve = BezierCurve(points, self.initial_time, self.final_time)
        return self * curve
    
    def __add__(self, curve : Self | Number) -> Self:
        if not isinstance(curve, BezierCurve):
            curve = self._number_to_curve(curve)
        self._assert_same_times(curve)
        if curve.degree > self.degree:
            self = self.elevate_degree(curve.degree)
        elif self.degree > curve.degree:
            curve = curve.elevate_degree(self.degree)
        points = self.points + curve.points
        return BezierCurve(points, self.initial_time, self.final_time)
    
    def __radd__(self, curve : Self | Number) -> Self:
        return self + curve
    
    def __sub__(self, curve : Self | Number) -> Self:
        return self + curve * (-1)
    
    def __rsub__(self, curve : Self | Number) -> Self:
        return self * (-1) + curve
    
    def  __neg__(self) -> Self:
        return 0 - self
    
    def derivative(self) -> Self:
        points = (self.points[1:] - self.points[:-1]) * (self.degree / self.duration)
        return BezierCurve(points, self.initial_time, self.final_time)

    def integral(self, initial_condition : np.ndarray | None = None) -> Self:
        points = self.points * self.duration / (self.degree + 1)
        points = np.vstack([np.zeros(self.dimension), points])
        points = np.cumsum(points, axis=0)
        if initial_condition is not None:
            points += initial_condition
        return BezierCurve(points, self.initial_time, self.final_time)
    
    def domain_split(self, time : float) -> Tuple[Self, Self]:
        if time < self.initial_time:
            raise ValueError("Split time must be larger than initial time.")
        if time > self.final_time:
            raise ValueError("Split time must be smaller than final time.")
        points = self.points
        points1 = np.zeros(self.points.shape, dtype=object)
        points2 = np.zeros(self.points.shape, dtype=object)
        c = (time - self.initial_time) / self.duration
        d = (self.final_time - time) / self.duration
        for i in range(self.degree):
            points1[i] = points[0]
            points2[-i-1] = points[-1]
            points = points[1:] * c + points[:-1] * d
        points1[-1] = points
        points2[0] = points
        curve1 = BezierCurve(points1, self.initial_time, time)
        curve2 = BezierCurve(points2, time, self.final_time)
        return curve1, curve2
    
    def l2_squared(self) -> float:
        """See (34) in Algorithms for polynomials in Bernstein form, by Farouky
        and Rajan"""
        a = 0
        for i in range(self.degree + 1):
            bi = binomial(self.degree, i)
            for j in range(i, self.degree + 1):
                bj = binomial(self.degree, j)
                bij = binomial(2 * self.degree, i + j)
                b = bi * bj / bij
                if j > i:
                    b *= 2
                a += b * self.points[i].dot(self.points[j])
        return self.duration * a / (2 * self.degree + 1)
    
    def integral_of_convex_function(self, f : Callable) -> float:
        c = self.duration / (self.degree + 1)
        return c * sum(f(point) for point in self.points)

    def plot_components(self, n : int = 51, legend : bool = True, **kwargs):
        import matplotlib.pyplot as plt
        colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        repeat = np.ceil(self.dimension / len(colors))
        colors = colors * int(repeat)
        t = np.linspace(self.initial_time, self.final_time, n)
        values = self(t)
        for i, value in enumerate(values.T):
            label = f"component {i}" if legend else None
            plt.plot(t, value, c=colors[i], label=label, **kwargs)
        if legend:
            plt.legend()

    def plot_trace_2d(self, n : int = 51, **kwargs):
        if self.dimension != 2:
            raise ValueError("Can only plot trace of 2d curves.")
        import matplotlib.pyplot as plt
        options = {"c": "b"}
        options.update(kwargs)
        t = np.linspace(self.initial_time, self.final_time, n)
        plt.plot(*self(t).T, **options)

    def scatter_points_2d(self, **kwargs):
        if self.dimension != 2:
            raise ValueError("Can only scatter points of 2d curves.")
        import matplotlib.pyplot as plt
        options = {"fc": "orange", "ec": "k", "zorder": 3}
        options.update(kwargs)
        plt.scatter(*self.points.T, **options)

    def plot_control_polytope_2d(self, **kwargs):
        if self.dimension != 2:
            raise ValueError("Can plot control polytope of 2d curves.")
        import matplotlib.pyplot as plt
        from matplotlib.patches import Polygon
        from scipy.spatial import ConvexHull
        options = {"fc": "lightcoral"}
        options.update(kwargs)
        hull = ConvexHull(self.points)
        ordered_points = hull.points[hull.vertices]
        poly = Polygon(ordered_points, **options)
        plt.gca().add_patch(poly)



import numpy as np
from typing import List, Callable, Self
from collections.abc import Iterable
from numbers import Number
from pybezier.bezier_curve import BezierCurve

class CompositeBezierCurve(object):

    def __init__(self, curves):
        initial_times = [curve.initial_time for curve in curves[1:]]
        final_times = [curve.final_time for curve in curves[:-1]]
        if not np.allclose(initial_times, final_times):
            raise ValueError("Initial and final times don't match.")
        dimensions = [curve.dimension for curve in curves]
        if len(set(dimensions)) != 1:
            raise ValueError("All the curves must have the same dimension.")
        self.curves = curves
        self.dimension = curves[0].dimension
        self.initial_time = curves[0].initial_time
        self.final_time = curves[-1].final_time
        self.duration = self.final_time - self.initial_time
        self.knot_times = [self.initial_time] + [curve.final_time for curve in curves]

    def curve_segment(self, time : float) -> int:
        segment = 0
        while self[segment].final_time < time:
            segment += 1
        return segment

    def __call__(self, time : float) -> np.ndarray:
        segment = self.curve_segment(time)
        return self[segment](time)

    def initial_point(self) -> np.ndarray:
        return self[0].initial_point()

    def final_point(self) -> np.ndarray:
        return self[-1].final_point()

    def __iter__(self) -> Iterable[BezierCurve]:
        return iter(self.curves)

    def __getitem__(self, i : int) -> BezierCurve:
        return self.curves[i]

    def __len__(self) -> int:
        return(len(self.curves))
    
    def __mul__(self, composite_curve : Self | Number) -> Self:
        if not isinstance(composite_curve, CompositeBezierCurve):
            composite_curve = self._number_to_composite_curve(composite_curve)
        curves = [curve_1 * curve_2 for curve_1, curve_2 in zip(self, composite_curve)]
        return CompositeBezierCurve(curves)

    def _number_to_composite_curve(self, n : Number) -> Self:
        curves = [curve._number_to_curve(n) for curve in self]
        return CompositeBezierCurve(curves)

    def __rmul__(self, composite_curve : Self | Number) -> Self:
        return self * composite_curve
    
    def elevate_degree(self, degree : int) -> Self:
        curves = [curve.elevate_degree(degree) for curve in self]
        return CompositeBezierCurve(curves)

    def __add__(self, composite_curve : Self | Number) -> Self:
        if not isinstance(composite_curve, CompositeBezierCurve):
            composite_curve = self._number_to_composite_curve(composite_curve)
        curves = [curve_1 + curve_2 for curve_1, curve_2 in zip(self, composite_curve)]
        return CompositeBezierCurve(curves)
    
    def __radd__(self, composite_curve : Self | Number) -> Self:
        return self + composite_curve
    
    def __sub__(self, composite_curve : Self | Number) -> Self:
        return self + composite_curve * (-1)
    
    def __rsub__(self, composite_curve : Self | Number) -> Self:
        return self * (-1) + composite_curve
    
    def  __neg__(self) -> Self:
        return 0 - self

    def knot_points(self) -> np.ndarray:
        # assumes that the curve is continuous
        knots = [curve.initial_point() for curve in self]
        knots.append(self.final_point())
        return np.array(knots)

    def durations(self) -> np.ndarray:
        return np.array([curve.duration for curve in self])

    def concatenate(self, composite_curve : Self) -> Self:
        shifted_curves = []
        for curve in composite_curve:
            initial_time = curve.initial_time + self.duration
            final_time = curve.final_time + self.duration
            shifted_curve = BezierCurve(curve.points, initial_time, final_time)
            shifted_curves.append(shifted_curve)
        return CompositeBezierCurve(self.curves + shifted_curves)

    def derivative(self) -> Self:
        return CompositeBezierCurve([curve.derivative() for curve in self])

    def integral(self, initial_condition : np.ndarray | None = None) -> Self:
        curves = []
        for curve in self:
            curves.append(curve.integral(initial_condition))
            initial_condition = curves[-1].final_point()
        return CompositeBezierCurve(curves)
    
    def l2_squared(self) -> float:
        return sum(curve.l2_squared() for curve in self)

    def integral_of_convex_function(self, f : Callable) -> float:
        return sum(curve.integral_of_convex_function(f) for curve in self)

    def plot_components(self, n : int = 51, legend : bool = True, **kwargs):
        for i, curve in enumerate(self):
            curve.plot_components(n, legend, **kwargs)
            if i == 0:
                legend = False
        
    def plot_trace_2d(self, **kwargs):
        for curve in self:
            curve.plot_trace_2d(**kwargs)

    def scatter_points_2d(self, **kwargs):
        for curve in self:
            curve.scatter_points_2d(**kwargs)

    def plot_control_polytopes_2d(self, **kwargs):
        for curve in self:
            curve.plot_control_polytope_2d(**kwargs)



import unittest
import numpy as np
from pybezier.bezier_curve import BezierCurve

class TestBezierCurve(unittest.TestCase):

    def setUp(self):
        np.random.seed(0)
        self.points = np.random.rand(5, 3)
        self.initial_time = .55
        self.final_time = 3.12
        self.curve = BezierCurve(self.points, self.initial_time, self.final_time)
        self.curve_1 = self.curve
        self.points_2 = np.random.rand(7, 3)
        self.curve_2 = BezierCurve(self.points_2, self.initial_time, self.final_time)
        self.time_samples = np.linspace(self.initial_time, self.final_time)

    def test_init(self):
        np.testing.assert_equal(self.curve.points, self.points)
        self.assertEqual(self.curve.initial_time, self.initial_time)
        self.assertEqual(self.curve.final_time, self.final_time)
        self.assertEqual(self.curve.duration, self.final_time - self.initial_time)
        self.assertEqual(self.curve.degree, self.points.shape[0] - 1)
        self.assertEqual(self.curve.dimension, self.points.shape[1])
        self.assertRaises(ValueError, BezierCurve, self.points, self.final_time, self.initial_time)

    def test_init_deafult(self):
        curve = BezierCurve(self.points)
        self.assertEqual(curve.initial_time, 0)
        self.assertEqual(curve.final_time, 1)

    def test_initial_final_point(self):
        np.testing.assert_array_almost_equal(self.curve(self.initial_time), self.curve.initial_point())
        np.testing.assert_array_almost_equal(self.curve(self.final_time), self.curve.final_point())

    def test_berstein(self):
        for time in self.time_samples:
            values = [self.curve._berstein(time, n) for n in range(self.curve.degree + 1)]
            # tests partition of unity
            self.assertTrue(min(values) >= 0)
            self.assertAlmostEqual(sum(values), 1)

    def test_assert_same_times(self):
        self.curve_1._assert_same_times(self.curve_2)
        curve_3 = BezierCurve(self.points)
        self.assertRaises(ValueError, self.curve_1._assert_same_times, curve_3)

    def test_call(self):
        points = np.array([[0, 0], [0, 1], [1, 1], [1, 0]])
        initial_time = 0
        final_time = 3
        curve = BezierCurve(points, initial_time, final_time)
        p1 = curve(1)
        p2 = curve(2)
        # tests convex-hull property
        np.testing.assert_array_less(p1, points[2])
        np.testing.assert_array_less(p2, points[2])
        np.testing.assert_array_less(points[0], p1)
        np.testing.assert_array_less(points[0], p2)
        # tests symmetry
        self.assertAlmostEqual(p1[0], 1 - p2[0])
        self.assertAlmostEqual(p1[1], p2[1])

    def test_scalar_mul(self):
        c = 3.66
        prod_1 = self.curve * c
        prod_2 = c * self.curve
        for time in self.time_samples:
            value = self.curve(time) * c
            np.testing.assert_array_almost_equal(prod_1(time), value)
            np.testing.assert_array_almost_equal(prod_2(time), value)
        
    def test_elementwise_mul(self):
        prod = self.curve_1 * self.curve_2
        for time in self.time_samples:
            np.testing.assert_array_almost_equal(prod(time), self.curve_1(time) * self.curve_2(time))

    def test_elevate_degree(self):
        curve = self.curve.elevate_degree(11)
        for time in self.time_samples:
            np.testing.assert_array_almost_equal(self.curve(time), curve(time))

    def test_scalar_add_sub(self):
        c = 3.66
        sum_1 = self.curve + c
        sum_2 = c + self.curve
        sub_1 = self.curve - c
        sub_2 = c - self.curve
        for time in self.time_samples:
            value = self.curve(time) + c
            np.testing.assert_array_almost_equal(sum_1(time), value)
            np.testing.assert_array_almost_equal(sum_2(time), value)
            value -= 2 * c
            np.testing.assert_array_almost_equal(sub_1(time), value)
            np.testing.assert_array_almost_equal(sub_2(time), -value)

    def test_elementwise_add_sub(self):
        sum = self.curve_1 + self.curve_2
        sub = self.curve_1 - self.curve_2
        for time in self.time_samples:
            value = self.curve_1(time) + self.curve_2(time)
            np.testing.assert_array_almost_equal(sum(time), value)
            value -= 2 * self.curve_2(time)
            np.testing.assert_array_almost_equal(sub(time), value)

    def test_neg(self):
        neg = - self.curve
        for time in self.time_samples:
            np.testing.assert_array_almost_equal(neg(time), -self.curve(time))

    def test_derivative(self):
        derivative = self.curve.derivative()
        time_step = 1e-6
        for time in np.linspace(self.initial_time, self.final_time - time_step):
            numerical_derivative = (self.curve(time + time_step) - self.curve(time)) / time_step
            np.testing.assert_array_almost_equal(derivative(time), numerical_derivative)

    def test_integral(self):
        initial_conditions = [None, np.ones(self.curve.dimension)]
        for initial_condition in initial_conditions:
            integral = self.curve.integral(initial_condition)
            value = integral(self.initial_time)
            target_value = 0 if initial_condition is None else initial_condition
            np.testing.assert_array_almost_equal(value, target_value)
            time_step = 1e-3
            for time in np.linspace(self.initial_time, self.final_time - time_step):
                value = self.curve(time + time_step / 2)
                target_value = (integral(time + time_step) - integral(time)) / time_step
                np.testing.assert_array_almost_equal(value, target_value)

    def test_domain_split(self):
        split_time = (self.initial_time + self.final_time) / 2
        curve_1, curve_2 = self.curve.domain_split(split_time)
        for time in self.time_samples:
            if time < split_time:
                np.testing.assert_array_almost_equal(self.curve(time), curve_1(time))
            elif time > split_time:
                np.testing.assert_array_almost_equal(self.curve(time), curve_2(time))

    def test_l2_squared(self):
        n_samples = 5000
        times = np.linspace(self.initial_time, self.final_time, n_samples)
        values = [self.curve(time).dot(self.curve(time)) for time in times]
        integral = np.trapezoid(values, times)
        self.assertAlmostEqual(self.curve.l2_squared(), integral)

    def test_integral_of_convex_function(self):
        f = lambda point: point.dot(point)
        value = self.curve.l2_squared()
        upper_bound = self.curve.integral_of_convex_function(f)
        self.assertTrue(value <= upper_bound)
        # upper bound for curve lenght is equal to distance of control points
        diff_points = self.points[:-1] - self.points[1:]
        value = sum(np.linalg.norm(diff_points, axis=1))
        derivative = self.curve.derivative()
        upper_bound = derivative.integral_of_convex_function(np.linalg.norm)
        self.assertAlmostEqual(value, upper_bound)
        
if __name__ == '__main__':
    unittest.main()


import unittest
import numpy as np
from pybezier.bezier_curve import BezierCurve
from pybezier.composite_bezier_curve import CompositeBezierCurve

class TestCompositeBezierCurve(unittest.TestCase):

    @staticmethod
    def random_composite_curve(dimension, n_curves, n_points):
        points = np.random.rand((n_points - 1) * n_curves + 1, dimension)
        curves = []
        for i in range(n_curves):
            start = n_points * i - i
            stop = n_points * (i + 1) - i
            points_i = points[start:stop]
            curve_i = BezierCurve(points_i, i, i + 1)
            curves.append(curve_i)
        return CompositeBezierCurve(curves)

    def setUp(self):
        np.random.seed(0)
        self.dimension = 3
        self.n_curves = 4
        self.n_points = 5
        self.composite_curve = self.random_composite_curve(self.dimension, self.n_curves, self.n_points)
        self.initial_time = 0
        self.final_time = self.n_curves
        self.time_samples = np.linspace(self.initial_time, self.final_time)
        self.composite_curve_1 = self.composite_curve
        self.composite_curve_2 = self.random_composite_curve(self.dimension, self.n_curves, self.n_points)

    def test_init(self):
        self.assertEqual(len(self.composite_curve.curves), self.n_curves)
        for curve in self.composite_curve.curves:
            self.assertEqual(curve.points.shape, (self.n_points, self.dimension))
        self.assertEqual(self.composite_curve.dimension, self.dimension)
        self.assertEqual(self.composite_curve.initial_time, 0)
        self.assertEqual(self.composite_curve.final_time, self.n_curves)
        self.assertEqual(self.composite_curve.duration, self.n_curves)
        self.assertEqual(self.composite_curve.knot_times, list(range(self.n_curves + 1)))

    def test_curve_segment(self):
        for time in self.time_samples[:-1]:
            self.assertEqual(self.composite_curve.curve_segment(time), int(time))
        self.assertEqual(self.composite_curve.curve_segment(self.final_time), self.n_curves - 1)

    def test_call(self):
        for time in self.time_samples:
            value = self.composite_curve(time)
            i = self.composite_curve.curve_segment(time)
            curve = self.composite_curve.curves[i]
            np.testing.assert_array_almost_equal(value, curve(time))

    def test_initial_final_point(self):
        initial_value = self.composite_curve(self.initial_time)
        initial_point = self.composite_curve.initial_point()
        np.testing.assert_array_almost_equal(initial_value, initial_point)
        final_value = self.composite_curve(self.final_time)
        final_point = self.composite_curve.final_point()
        np.testing.assert_array_almost_equal(final_value, final_point)

    def test_iter(self):
        for curve in self.composite_curve:
            self.assertEqual(curve.duration, 1)

    def test_getitem(self):
        for i in range(self.n_curves):
            curve = self.composite_curve[i]
            self.assertEqual(curve.duration, 1)

    def test_len(self):
        self.assertEqual(len(self.composite_curve), self.n_curves)

    def test_scalar_mul(self):
        c = 3.66
        prod_1 = self.composite_curve * c
        prod_2 = c * self.composite_curve
        for time in self.time_samples:
            value = self.composite_curve(time) * c
            np.testing.assert_array_almost_equal(prod_1(time), value)
            np.testing.assert_array_almost_equal(prod_2(time), value)
        
    def test_elementwise_mul(self):
        prod = self.composite_curve_1 * self.composite_curve_2
        for time in self.time_samples:
            value = self.composite_curve_1(time) * self.composite_curve_2(time)
            np.testing.assert_array_almost_equal(prod(time), value)

    def test_elevate_degree(self):
        composite_curve = self.composite_curve.elevate_degree(11)
        for time in self.time_samples:
            np.testing.assert_array_almost_equal(self.composite_curve(time), composite_curve(time))

    def test_scalar_add_sub(self):
        c = 3.66
        sum_1 = self.composite_curve + c
        sum_2 = c + self.composite_curve
        sub_1 = self.composite_curve - c
        sub_2 = c - self.composite_curve
        for time in self.time_samples:
            value = self.composite_curve(time) + c
            np.testing.assert_array_almost_equal(sum_1(time), value)
            np.testing.assert_array_almost_equal(sum_2(time), value)
            value -= 2 * c
            np.testing.assert_array_almost_equal(sub_1(time), value)
            np.testing.assert_array_almost_equal(sub_2(time), -value)

    def test_elementwise_add_sub(self):
        sum = self.composite_curve_1 + self.composite_curve_2
        sub = self.composite_curve_1 - self.composite_curve_2
        for time in self.time_samples:
            value = self.composite_curve_1(time) + self.composite_curve_2(time)
            np.testing.assert_array_almost_equal(sum(time), value)
            value -= 2 * self.composite_curve_2(time)
            np.testing.assert_array_almost_equal(sub(time), value)

    def test_neg(self):
        neg = - self.composite_curve
        for time in self.time_samples:
            np.testing.assert_array_almost_equal(neg(time), -self.composite_curve(time))

    def test_derivative(self):
        derivative = self.composite_curve.derivative()
        time_step = 1e-7
        for time in np.linspace(self.initial_time, self.final_time - time_step):
            numerical_derivative = (self.composite_curve(time + time_step) - self.composite_curve(time)) / time_step
            np.testing.assert_array_almost_equal(derivative(time), numerical_derivative)

    def test_integral(self):
        initial_conditions = [None, np.ones(self.composite_curve.dimension)]
        for initial_condition in initial_conditions:
            integral = self.composite_curve.integral(initial_condition)
            value = integral(self.initial_time)
            target_value = 0 if initial_condition is None else initial_condition
            np.testing.assert_array_almost_equal(value, target_value)
            time_step = 1e-3
            for time in np.linspace(self.initial_time, self.final_time - time_step):
                value = self.composite_curve(time + time_step / 2)
                target_value = (integral(time + time_step) - integral(time)) / time_step
                np.testing.assert_array_almost_equal(value, target_value)

    def test_knot_points(self):
        for i, point in enumerate(self.composite_curve.knot_points()):
            np.testing.assert_array_almost_equal(point, self.composite_curve(i))

    def test_durations(self):
        durations = list(self.composite_curve.durations())
        self.assertEqual(durations, [1] * self.n_curves)

    def test_concatenate(self):
        conc = self.composite_curve_1.concatenate(self.composite_curve_2)
        for time in self.time_samples * 2:
            if time < self.final_time:
                value = self.composite_curve_1(time)
                np.testing.assert_array_almost_equal(conc(time), value)
            elif time > self.final_time:
                value = self.composite_curve_2(time - self.final_time)
                np.testing.assert_array_almost_equal(conc(time), value)    
                
    def test_l2_squared(self):
        n_samples = 5000
        times = np.linspace(self.initial_time, self.final_time, n_samples)
        squared_norm = lambda time: np.linalg.norm(self.composite_curve(time)) ** 2
        values = [squared_norm(time) for time in times]
        integral = np.trapezoid(values, times)
        self.assertAlmostEqual(self.composite_curve.l2_squared(), integral, places=4)

    def test_integral_of_convex_function(self):
        f = lambda point: point.dot(point)
        value = self.composite_curve.l2_squared()
        upper_bound = self.composite_curve.integral_of_convex_function(f)
        self.assertTrue(value <= upper_bound)
        # upper bound for curve lenght is equal to distance of control points
        diff_points = np.vstack([curve.points[:-1] - curve.points[1:] for curve in self.composite_curve])
        value = sum(np.linalg.norm(diff_points, axis=1))
        derivative = self.composite_curve.derivative()
        upper_bound = derivative.integral_of_convex_function(np.linalg.norm)
        self.assertAlmostEqual(value, upper_bound)

if __name__ == '__main__':
    unittest.main()


import unittest
from pybezier.binomial import binomial

class TestBinomial(unittest.TestCase):

    def test_valid_inputs(self):
        self.assertEqual(binomial(4, 2), 6)
        self.assertEqual(binomial(8, 3), 56)
        self.assertEqual(binomial(8, 8), 1)
        self.assertEqual(binomial(9, 5), 126)
        self.assertEqual(binomial(9, 7), 36)
        self.assertEqual(binomial(11, 0), 1)

    def test_invalid_inputs(self):
        self.assertRaises(ValueError, binomial, 4, .5)
        self.assertRaises(ValueError, binomial, 4.5, 2)
        self.assertRaises(ValueError, binomial, 4, 5)
        self.assertRaises(ValueError, binomial, 4, -1)

if __name__ == '__main__':
    unittest.main()


from setuptools import setup

setup(
    name='pybezier',
    version='0.1.0',
    description='A simple Python package to manipulate Bézier curves',
    url='https://github.com/TobiaMarcucci/pybezier',
    author='Tobia Marcucci',
    author_email='tobiam@mit.edu',
    license='MIT',
    packages=['pybezier'],
    install_requires=['numpy'],
)
